#!/usr/bin/env python
#coding: utf-8

from milenasimple import *
import milenasimple

class Milena(milenasimple._milenka):
    """
    Interfejs wyższego poziomu do procesora języka Milena
    """
    def __init__(self,flags=None,ivoice=None,
                 langs=None,
                 themes=None,
                 dic=None,
                 phraser=None,
                 mbrola=None,
                 ivocoding='utf-8'):
        self._enc=ivocoding.lower()
        self._flags=0
        self._paragraph=None
        self._last_partype=-1
        self._ivoice=ivoice
        
        minit={}
        if flags:
            minit['flags']=flags
            self._flags=flags
        if ivoice:
            minit['ivoice']=ivoice
        milenasimple._milenka.__init__(self,**minit)
        if langs:
            if isinstance(langs,str):
                langs=[langs]
            for lang in langs:
                self.loadLang(lang)
        if themes:
            if isinstance(themes,str):
                themes=[themes]
            for theme in themes:
                self.loadTheme(theme)
        if dic:
            if isinstance(dic,str):
                dic=[dic]
            for a in dic:
                self.loadDic(a)
        if phraser:
            if isinstance(phraser,str):
                phraser=[phraser]
            for a in phraser:
                self.loadPhraser(a)
        if mbrola:
            if isinstance(mbrola,str):
                mbrola=[mbrola]
            for a in mbrola:
                self.loadMbrola(a)
        
    def getPhrase(self,text,breath=0):
        """getPhrase(text,breath=0)
Pobiera frazę dla Mbroli z wprowadzonego tekstu. Jeśli breath != 0,
wyjście poprzedzane jest fonemem pauzy odpowiadającej długości
oddechu.
Zwracana wartość: krotka(fonemy,typ_frazy,reszta_tekstu,przeczytany_tekst)
lub None jeśli koniec tekstu.
Jeśli parametr text jest unikodem, trzecia i czwarta wartość krotki
wynikowej będą również unikodem.
"""
        if not isinstance(text,unicode):
            return milenasimple._milenka.getPhrase(self,text,breath)
        text=text.encode('utf-8')
        d=milenasimple._milenka.getPhrase(self,text,breath)
        if d:
            d=(d[0],d[1],d[2].decode('utf-8') if d[2] is not None else None,d[3].decode('utf-8'))
        return d
    
    def getSentence(self,text):
        """getSentence(text)
Pobiera fonetyczną wersję zdania z wprowadzonego tekstu.
Zwracana wartość: krotka(fonemy,typ_frazy,reszta_tekstu,przeczytany_tekst)
lub None jeśli koniec tekstu.
Jeśli parametr text jest unikodem, trzecia i czwarta wartość krotki
wynikowej będą również unikodem.
Kodowanie fonemów określa parametr ivocoding inicjalizatora klasy,
dopuszczalne wartości to: 'utf-8', 'iso-8859-2', 'unicode'.
"""
        is_uni=False
        if isinstance(text,unicode):
            is_uni=True
            text=text.encode('utf-8')
        d=milenasimple._milenka.getSentence(self,text)
        if not d:
            return None
        if self._enc == 'utf-8':
            d=(d[0].decode('iso-8859-2').encode('utf-8'),d[1],d[2],d[3])
        elif self._enc == 'unicode':
            d=(d[0].decode('iso-8859-2'),d[1],d[2],d[3])
        if is_uni:
            d=(d[0],d[1],d[2].decode('utf-8'),d[3].decode('utf-8'))
        return d
    
    def getParType(self,text):
        """getParType(text)
Ustala typ akapitu na:
PARTYPE_EMPTY - dla akapitu pustego lub nie zawierającego
   wymawialnego tekstu
PARTYPE_DIALOG - dla akapitu będącego kwestia dialogu
PARTYPE_NORMAL - dla normalnego akapitu.
"""
        if not isinstance(text,unicode):
            text=text.decode('utf-8')
        #print "GOT", '%x'%ord(text[0]), text
        text=text.strip()
        if text[0] in u"-\u2012\u2013\u2014\u2015" and len(text) < 2:
        #    print "PEM"
            return PARTYPE_EMPTY
        dialog=PARTYPE_NORMAL
        if text[0] == '-' and text[1].isdigit():
            return PARTYPE_NORMAL
        if text[0] in u"-\u2012\u2013\u2014\u2015":
            text=text[1:]
            dialog=PARTYPE_DIALOG
        #print "IAN",self.isAlNum(ord(text[0]))
        for a in text:
            if self.isAlNum(ord(a)):
                return dialog
        return PARTYPE_EMPTY
    
    def feed(self,text=None):
        """feed(text=None)
Bez parametru - ustawia wewnętrzny marker ostatniego akapitu
na początek treści i zwraca None.
Z parametrem - sprawdza typ akapitu, na podstawie markera
ostatniego akapitu ustala długość oddechu i ustawia marker
ostatniego akapitu na typ bieżącego.
Zwraca krotkę (typ_akapitu, oddech)
"""
        if text is None:
            self._paragraph=None
            self._last_partype=-1
            return None
        pt=self.getParType(text)
        breath=0
        if pt == PARTYPE_EMPTY:
            self._paragraph=None
        else:
            self._paragraph=text
            if self._last_partype >= 0:
                if self._last_partype == PARTYPE_EMPTY:
                    breath=BREATH_LONG
                elif self._last_partype == PARTYPE_DIALOG:
                    breath=BREATH_DIALOG if pt == PARTYPE_DIALOG else BREATH_POSTDIAL
                else:
                    breath=BREATH_NORMAL if pt == PARTYPE_NORMAL else BREATH_PREDIAL
        self._last_partype = PARTYPE_EMPTY
        self._breath=breath
        return (self._last_partype,breath)
    
    def consume(self):
        """consume()
Pobiera kolejne zdanie (w trybie Ivony) lub frazę (w trybie Mbroli)
z tekstu podanego w metodzie feed().
Zwraca krotkę(fraza_fonetyczna, fraza_tekstowa, typ)
gdzie:
  fraza_fonetyczna to tekst dla Ivony lub fonemu Mbroli
  fraza_tekstowa to pobrany tekst
  typ to typ frazy (np. PHRASE_TYPE_DOT dla zdania zakończonego kropką).
Fraza tekstowa będzie tego samego typu co parametr feed
(unicode lub str kodowany utf-8)
Jeśli nie ma już więcej tekstu, zwraca None
"""
        if not self._paragraph:
            return None
        breath=self._breath
        self._breath=0
        if self._ivoice:
            d=self.getSentence(self._paragraph)
        else:
            d=self.getPhrase(self._paragraph,breath)
        if not d:
            self._paragraph=None
            return None
        self._paragraph=d[2]
        return (d[0],d[3],d[1])
